// --------------- SINGLE-CHANNEL ROUTINES ---------------
// wavelet image compression
// encodes channels (which necessarily have to have
// power of 2 dimensions) using a wavelet-transform and
// bit-plane run-length coding into a progressive bit-stream
// (c) 2001-2003 Daniel Vollmer (maven@maven.de)

// 2.7	-	removed MMX optimisations for wavelet transforms and made code even faster ;)
//		-	removed unused MaxBits parameter from wv_init_channel
//		-	changed bitstream format (order in which bits are written) and removed writing
//			unneccessary zeros at the end of each block
//		-	changed yuv transform slightly (Cr / CB are now centred around 0, not 128), as we're writing
//			the sign in any case
//		-	changed colorspace conversion to be in-place
//		-	fixed bug in raw_load if file was too small
//		-	misc optimisations to bit-files
//		-	added wv_ prefix to log2i, mse_to_psnr, psnr_to_mse
//		-	changed the # of iterations for the multi-channel size selector (now bails out earlier)
//		-	new function to return the header of an image (wv_read_header), changed layout of t_wv_dchannels
//		-	changed decompression to work for (hopefully) all invalid data w/o overwriting
//			anything in memory
//		-	wv_init_decode_channels now accepts an extra reduction parameter (return a scaled down version of image)
//			(see -dr parameter in wako.exe)
// 2.6	-	reordering now clears unused coefficients as well
//		-	init_channel now accepts a NEGATIVE Lossless parameter to indicate that the
//			(faster) approximate error estimation is to be used
//		-	optimised the wavelet-transforms (including MMX versions)
//		-	wrote seperate size-counter for rice-encoding
// 2.5	-	changed reordering again slightly, now blocks (VD, HD, DD) are transmitted
//			completely interleaved, reordering lowest level unnecessarily fixed
//		-	fixed bug with blocks consisting of all 0s
// 2.1	-	file slightly smaller -> reordered '0' coefficients to the end, incompatible

/* 	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA	02111-1307	USA
*/

//#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <time.h>
#include "wavelet.h"

#define wv_MIN_HEADER_VERSION 1
#define wv_CUR_HEADER_VERSION 1
#define wv_MIN_BITSTREAM_VERSION 1
#define wv_CUR_BITSTREAM_VERSION 1

#define DETAIL(ptr, pitch) ((ptr)[0] -= ((ptr)[-(pitch)] + (ptr)[(pitch)]) / 2)
#define DETAIL_LAST(ptr, pitch) ((ptr)[0] -= (ptr)[-(pitch)])
#define AVERAGE(ptr, pitch) ((ptr)[-(pitch)] += ((ptr)[0] + (ptr)[-2 * (pitch)]) / 4)
#define AVERAGE_FIRST(ptr, pitch) ((ptr)[-(pitch)] += (ptr)[0] / 2)

static void wavelet_row(wv_pel *dst, const int pitch, const int count)
{ // biorthogonal (2,2) Cohen-Daubechies-Feauveau (lifting)
	int i;
	wv_pel *ptr;
	
	ptr = dst + pitch; // odd
	
	DETAIL(ptr, pitch);
	AVERAGE_FIRST(ptr, pitch);
	for (i = count - 2; i > 0; i--)
	{
		ptr += 2 * pitch;
		DETAIL(ptr, pitch);
		AVERAGE(ptr, pitch);
	}
	ptr += 2 * pitch;
	DETAIL_LAST(ptr, pitch);
	AVERAGE(ptr, pitch);
}

static void wavelet_even_no_row_first(wv_pel *dst, const int pitch, const int rowpitch, const int count)
{ // biorthogonal (2,2) Cohen-Daubechies-Feauveau (lifting)
	int i;
	wv_pel *rowptr;
	
	rowptr = dst - rowpitch;
	
	for (i = 2 * count; i > 0; i--)
	{
		// column (first)
		DETAIL(rowptr, rowpitch);
		AVERAGE_FIRST(rowptr, rowpitch);
		rowptr += pitch;
	}
}

static void wavelet_even_no_row(wv_pel *dst, const int pitch, const int rowpitch, const int count)
{ // biorthogonal (2,2) Cohen-Daubechies-Feauveau (lifting)
	int i;
	wv_pel *rowptr;
	
	rowptr = dst - rowpitch;
	
	for (i = 2 * count; i > 0; i--)
	{
		// column (generic)
		DETAIL(rowptr, rowpitch);
		AVERAGE(rowptr, rowpitch);
		rowptr += pitch;
	}
}

static void wavelet_no_row_last(wv_pel *dst, const int pitch, const int rowpitch, const int count)
{ // biorthogonal (2,2) Cohen-Daubechies-Feauveau (lifting)
	int i;
	wv_pel *rowptr;
	
	rowptr = dst;
	
	for (i = 2 * count; i > 0; i--)
	{
		// column (last)
		DETAIL_LAST(rowptr, rowpitch);
		AVERAGE(rowptr, rowpitch);
		rowptr += pitch;
	}
}

// dst must contain the original (untransformed) image
// also width & height must be divisible by 2^(levels+1)
static void wavelet_transform(wv_pel *dst, const int width, const int height, const int levels)
{
	int l;

	for (l = 0; l < levels; l++)
	{
		int wmid = width >> (l + 1), hmid = height >> (l + 1);
		int pitch = 1 << l, rowpitch = width << l;
		int i;

		// precalculate the first three rows
		wavelet_row(dst, pitch, wmid);
		wavelet_row(dst + pitch * width, pitch, wmid);
		wavelet_row(dst + 2 * pitch * width, pitch, wmid);
		// do the third row (and the 1st two columns)
		wavelet_even_no_row_first(dst + 2 * pitch * width, pitch, rowpitch, wmid);
		for (i = 3 * pitch; i < height - pitch; i += 2 * pitch)
		{
			wavelet_row(dst + i * width, pitch, wmid); // odd row
			wavelet_row(dst + (i + pitch) * width, pitch, wmid); // even row
			wavelet_even_no_row(dst + (i + pitch) * width, pitch, rowpitch, wmid); // even row + previous two columns
		}
		wavelet_row(dst + i * width, pitch, wmid); // last (odd) row
		wavelet_no_row_last(dst + i * width, pitch, rowpitch, wmid); // and the last two columns
	}
}


#define INV_AVERAGE_FIRST(ptr, pitch) ((ptr)[0] -= (ptr)[(pitch)] / 2)
#define INV_AVERAGE(ptr, pitch) ((ptr)[0] -= ((ptr)[-(pitch)] + (ptr)[(pitch)]) / 4)
#define INV_DETAIL(ptr, pitch) ((ptr)[-(pitch)] += ((ptr)[-2 * (pitch)] + (ptr)[0]) / 2)
#define INV_DETAIL_LAST(ptr, pitch) ((ptr)[(pitch)] += (ptr)[0])

static void inverse_wavelet_row(wv_pel *dst, const int pitch, const int count)
{
	int i;
	wv_pel *ptr;
	
	ptr = dst; // even (average)
	
	INV_AVERAGE_FIRST(ptr, pitch);
	for (i = count - 1; i > 0; i--)
	{
		ptr += 2 * pitch;
		INV_AVERAGE(ptr, pitch);
		INV_DETAIL(ptr, pitch);
	}
	INV_DETAIL_LAST(ptr, pitch);
}

static void inverse_wavelet_no_row_first(wv_pel *dst, const int pitch, const int rowpitch, const int count)
{
	int i;
	wv_pel *rowptr;
	
	rowptr = dst;
	
	for (i = 2 * count; i > 0; i--)
	{
		// column (first)
		INV_AVERAGE_FIRST(rowptr, rowpitch);
		rowptr += pitch;
	}
}

static void inverse_wavelet_no_row_evenm1(wv_pel *dst, const int pitch, const int rowpitch, const int count)
{
	int i;
	wv_pel *rowptr;
	
	// column
	rowptr = dst;
	
	for (i = 2 * count; i > 0; i--)
	{
		// column
		INV_AVERAGE(rowptr, rowpitch);
		INV_DETAIL(rowptr, rowpitch);
		rowptr += pitch;
	}
}

static void inverse_wavelet_no_row_last(wv_pel *dst, const int pitch, const int rowpitch, const int count)
{
	int i;
	wv_pel *rowptr;
	
	// column
	rowptr = dst;
	
	for (i = 2 * count; i > 0; i--)
	{
		// column
		INV_DETAIL_LAST(rowptr, rowpitch);
		rowptr += pitch;
	}
}


// dst must contain the wavelet transform of the image
// also width & height must be divisible by 2^(levels+1)
static void inverse_wavelet_transform(wv_pel *dst, const int width, const int height, const int levels)
{
	int l;

	for (l = levels - 1; l >= 0; l--)
	{
		int wmid = width >> (l + 1), hmid = height >> (l + 1);
		int pitch = 1 << l, rowpitch = width << l;
		int i;

		// preculate first row (w/ vertical transform)
		inverse_wavelet_no_row_first(dst, pitch, rowpitch, wmid);
		for (i = 2 * pitch; i < height; i += 2 * pitch)
		{
			inverse_wavelet_no_row_evenm1(dst + i * width, pitch, rowpitch, wmid);
			inverse_wavelet_row(dst + (i - 2 * pitch) * width, pitch, wmid);
			inverse_wavelet_row(dst + (i - pitch) * width, pitch, wmid);
		}
		inverse_wavelet_no_row_last(dst + (i - 2 * pitch) * width, pitch, rowpitch, wmid);
		inverse_wavelet_row(dst + (i - 2 * pitch) * width, pitch, wmid);
		inverse_wavelet_row(dst + (i - pitch) * width, pitch, wmid);
	}
}


/*static void quantize(wv_pel* dst, const int num, const int quant)
{
	if (quant > 1)
	{
		int i;

		for (i = 0; i < num; i++)
			*dst++ /= quant;
	}
}*/


/*int dead_zone_dequant(const int val, const int fac)
{ // has been multiplied by 10 to avoid floating point calculations entirely
	if (val > 0)
		return ((val * 10 + 5) * fac) / 10; // (val + 0.5) * fac
	else if (val < 0)
		return ((val * 10 - 5) * fac) / 10; // (val - 0.5) * fac
	else 
		return 0;
}

static void dequantize(wv_pel* dst, const int num, const int quant)
{
	if (quant > 1)
	{
		int i;

		// start dequantizing w/ T1
		for (i = 0; i < num; i++)
			*dst++ = (wv_pel)dead_zone_dequant(*dst, quant);
	}
}*/

#define dead_zone_dequant(v, f) (((v) > 0) ? ((((v) << 1) + 1) * (f)) >> 1 : ((v) < 0) ? -((((-(v) << 1) + 1) * (f)) >> 1) : 0)

// for the why and how of the function see "wavelet_analysis.nb". why it isn't exact (which i think it should be), ... i don't know!
static float estimate_error(const wv_pel *a, const int width, const int owidth, const int height, const int oheight, const int level, const int quant)
{
	if (quant > 1)
	{
		if (a)
		{
			int block_size_x = width >> level; // how big is a block ("level") of coefficients
			int block_size_y = height >> level; // in x and y?
			int unused_size_x = (width - owidth) >> level;
			int unused_size_y = (height - oheight) >> level;
			int pitch = 1 << level;
			int HDofs = (pitch >> 1), VDofs = (pitch >> 1) * width;
			int y, x;
			double sum2 = 0.0;

			#define OFS(x, y) ((y) * pitch * width + (x) * pitch)
			#define HD(ofs) a[(ofs) + HDofs]
			#define VD(ofs) a[(ofs) + VDofs]
			#define DD(ofs) a[(ofs) + HDofs + VDofs]
			#define ERR(v) ((v) - dead_zone_dequant((v) / quant, quant))
			
			for (y = 0; y < block_size_y - unused_size_y; y++)
			{
				#define HDI 0
				#define VDI 1
				#define DDI 2
				#define LI 0
				#define CI 1
				#define RI 2
				
				int yli = max(0, y - 1), yri = min(block_size_y - 1, y + 1); // y index left, right
				int errs[DDI + 1][RI + 1][RI + 1];
				int ofs;

				// we copy the error in the coefficients into this array and then only update the ones we need (after shifting them)
				// this way then inner loop only references fixed offsets into errs
				ofs = OFS(0, yli);
				errs[HDI][LI][CI] = errs[HDI][LI][RI] = ERR(HD(ofs));
				errs[VDI][LI][CI] = errs[VDI][LI][RI] = ERR(VD(ofs));
				errs[DDI][LI][CI] = errs[DDI][LI][RI] = ERR(DD(ofs));
				ofs = OFS(0, y);
				errs[HDI][CI][CI] = errs[HDI][CI][RI] = ERR(HD(ofs));
				errs[VDI][CI][CI] = errs[VDI][CI][RI] = ERR(VD(ofs));
				errs[DDI][CI][CI] = errs[DDI][CI][RI] = ERR(DD(ofs));
				ofs = OFS(0, yri);
				errs[HDI][RI][CI] = errs[HDI][RI][RI] = ERR(HD(ofs));
				errs[VDI][RI][CI] = errs[VDI][RI][RI] = ERR(VD(ofs));
				errs[DDI][RI][CI] = errs[DDI][RI][RI] = ERR(DD(ofs));
						
				for (x = 0; x < block_size_x - unused_size_x; x++)
				{
					int xri = min(block_size_x - 1, x + 1); // right
					double diff;

					// shift
					errs[HDI][LI][LI] = errs[HDI][LI][CI];
					errs[HDI][CI][LI] = errs[HDI][CI][CI];
					errs[HDI][RI][LI] = errs[HDI][RI][CI];

					errs[HDI][LI][CI] = errs[HDI][LI][RI];
					errs[HDI][CI][CI] = errs[HDI][CI][RI];
					errs[HDI][RI][CI] = errs[HDI][RI][RI];
					// shift
					errs[VDI][LI][LI] = errs[VDI][LI][CI];
					errs[VDI][CI][LI] = errs[VDI][CI][CI];
					errs[VDI][RI][LI] = errs[VDI][RI][CI];

					errs[VDI][LI][CI] = errs[VDI][LI][RI];
					errs[VDI][CI][CI] = errs[VDI][CI][RI];
					errs[VDI][RI][CI] = errs[VDI][RI][RI];
					// shift
					errs[DDI][LI][LI] = errs[DDI][LI][CI];
					errs[DDI][CI][LI] = errs[DDI][CI][CI];
					errs[DDI][RI][LI] = errs[DDI][RI][CI];

					errs[DDI][LI][CI] = errs[DDI][LI][RI];
					errs[DDI][CI][CI] = errs[DDI][CI][RI];
					errs[DDI][RI][CI] = errs[DDI][RI][RI];
					
					// fill
					ofs = OFS(xri, yli);
					errs[HDI][LI][RI] = ERR(HD(ofs));
					errs[VDI][LI][RI] = ERR(VD(ofs));
					errs[DDI][LI][RI] = ERR(DD(ofs));
					ofs = OFS(xri, y);
					errs[HDI][CI][RI] = ERR(HD(ofs));
					errs[VDI][CI][RI] = ERR(VD(ofs));
					errs[DDI][CI][RI] = ERR(DD(ofs));
					ofs = OFS(xri, yri);
					errs[HDI][RI][RI] = ERR(HD(ofs));
					errs[VDI][RI][RI] = ERR(VD(ofs));
					errs[DDI][RI][RI] = ERR(DD(ofs));
					

					// x even, y even
					diff = (errs[DDI][LI][LI] + errs[DDI][LI][CI] + errs[DDI][CI][LI] + errs[DDI][CI][CI]
						- 4 * errs[HDI][CI][LI] - 4 * errs[HDI][CI][CI]
						- 4 * errs[VDI][LI][CI] - 4 * errs[VDI][CI][CI]) / 16.0;
					sum2 += diff * diff;
					
					// x odd, y even
					diff = (errs[DDI][LI][LI] + errs[DDI][LI][RI] - 6 * errs[DDI][LI][CI] + errs[DDI][CI][LI] + errs[DDI][CI][RI] - 6 * errs[DDI][CI][CI]
						- 4 * errs[HDI][CI][LI] - 4 * errs[HDI][CI][RI] + 24 * errs[HDI][CI][CI]
						- 4 * errs[VDI][LI][RI] - 4 * errs[VDI][LI][CI] - 4 * errs[VDI][CI][RI] - 4 * errs[VDI][CI][CI]) / 32.0;
					sum2 += diff * diff;

					// x even, y odd
					diff = (errs[DDI][LI][LI] + errs[DDI][LI][CI] + errs[DDI][RI][LI] + errs[DDI][RI][CI] - 6 * errs[DDI][CI][LI] - 6 * errs[DDI][CI][CI]
						- 4 * errs[HDI][RI][LI] - 4 * errs[HDI][RI][CI] - 4 * errs[HDI][CI][LI] - 4 * errs[HDI][CI][CI]
						- 4 * errs[VDI][LI][CI] - 4 * errs[VDI][RI][CI] + 24 * errs[VDI][CI][CI]) / 32.0;
					sum2 += diff * diff;
					
					// x odd, y odd
					diff = (errs[DDI][LI][LI] + errs[DDI][LI][RI] - 6 * errs[DDI][LI][CI] + errs[DDI][RI][LI] + errs[DDI][RI][RI] + - 6 * errs[DDI][RI][CI] - 6 * errs[DDI][CI][LI] - 6 * errs[DDI][CI][RI] + 36 * errs[DDI][CI][CI]
						- 4 * errs[HDI][RI][LI] - 4 * errs[HDI][RI][RI] + 24 * errs[HDI][RI][CI] - 4 * errs[HDI][CI][LI] - 4 * errs[HDI][CI][RI] + 24 * errs[HDI][CI][CI]
						- 4 * errs[VDI][LI][RI] - 4 * errs[VDI][LI][CI] - 4 * errs[VDI][RI][RI] - 4 * errs[VDI][LI][CI] + 24 * errs[VDI][CI][RI] + 24 * errs[VDI][CI][CI]) / 64.0;
					sum2 += diff * diff;
				}
				#undef HDI
				#undef VDI
				#undef DDI
				#undef LI
				#undef CI
				#undef RI
			}
			#undef OFS
			#undef HD
			#undef VD
			#undef DD
			#undef ERR
			return (float)(sum2 / (float)((block_size_x - unused_size_x) * (block_size_y - unused_size_y) * 4));
		}
		else // estimate errors
		{
			int numq = quant / 2 - 1;
			double est_error;

			est_error = (double)quant / 4.0; // 1 / q * (1 / 2)^2

			// Sum[i^2, {i, n}] = 1/6 n (1 + n) (1 + 2n)
			est_error += (double)(numq * (1 + numq) * (1 + 2 * numq)) / (3.0 * (double)quant);

			return (float)est_error;
		}
	}
	return 0.0f;
}



//----------------------------------------------------------------------------
// wv_log2i		returns the integer logarithm to the base 2 of a given integer
//				(i.e. the number of bits needed to encode that number)
// max_val		number whose log is wanted (log2i(0) = 0)
//----------------------------------------------------------------------------
WAVELET_DLL_API int WAVELET_DLL_CC wv_log2i(int max_val)
{
	int i;

	for (i = 0; max_val > 0; max_val >>= 1, i++) ;
	return i;
}


// this converts a reorder-index into an (x,y) coordinate to fetch from
// using bit-interleaving (fractal z-curve)
static void get_zig_zag_idx(const int idx, const int maxx, const int maxy, int* xx, int* yy)
{
	int xbits = wv_log2i(maxx) - 1;
	int ybits = wv_log2i(maxy) - 1;
	int mask;

	mask = 1 << (xbits + ybits - 1);
	*xx = *yy = 0;
	if (ybits >= xbits)
	{
		while (ybits > 0)
		{
			*yy <<= 1;
			*yy |= (idx & mask) > 0;
			mask >>= 1;
			ybits--;
			if (xbits > 0)
			{
				*xx <<= 1;
				*xx |= (idx & mask) > 0;
				mask >>= 1;
				xbits--;
			}
		}
	}
	else
	{
		while (xbits > 0)
		{
			*xx <<= 1;
			*xx |= (idx & mask) > 0;
			mask >>= 1;
			xbits--;
			if (ybits > 0)
			{
				*yy <<= 1;
				*yy |= (idx & mask) > 0;
				mask >>= 1;
				ybits--;
			}
		}
	}
//	printf("x: %i y: %i -> idx: %i x: %i, y: %i\n", x, y, i, *xx, *yy);
}

static int* make_single_table(const int width, const int owidth, const int height, const int oheight, const int l, const int r)
{
	int block_size_x = width >> (l + r); // how big is a block ("level") of coefficients
	int block_size_y = height >> (l + r); // in x and y?
	int *tab = malloc((1 + block_size_x * block_size_y) * sizeof *tab);
	int *cur_table = tab + 1;
	int unused_size_x = (width - owidth) >> (l + r), unused_size_y = (height - oheight) >> (l + r);
	int i, j, pitch;
			
	pitch = 1 << l;
	tab[0] = 0; // number of non-zeroed entries
	for (i = 0; i < block_size_x * block_size_y; i++)
	{
		int k;

		get_zig_zag_idx(i, block_size_x, block_size_y, &j, &k);
		if (j <= (block_size_x - unused_size_x) && k <= (block_size_y - unused_size_y))
		{
			*cur_table++ = k * pitch * (width >> r) + j * pitch;
			tab[0]++; // increase number of non-zeroed entries
		}
	}
	// now add the unused ones (which should be all zero, anyway)
	for (i = 0; i < min(block_size_y, block_size_y - unused_size_y + 1); i++)
		for (j = block_size_x - unused_size_x + 1; j < block_size_x; j++)
			*cur_table++ = i * pitch * (width >> r) + j * pitch;
	for (i = block_size_y - unused_size_y + 1; i < block_size_y; i++)
		for (j = 0; j < block_size_x; j++)
			*cur_table++ = i * pitch * (width >> r) + j * pitch;
	return tab;
}

static int** init_reorder_tables(const int width, const int owidth, const int height, const int oheight, const int levels, const int reduce)
{
	int** tables = NULL;

	if (width > 2 && height > 2)
	{
		int l;
		
		tables = malloc(levels * sizeof *tables);
		for (l = 1; l <= levels; l++)
			tables[levels - l] = make_single_table(width, owidth, height, oheight, l, reduce);
	}
	return tables;
}


static void free_reorder_tables(int** tables, const int levels)
{
	if (tables)
	{
		int i;

		for (i = 0; i < levels; i++)
			free(tables[i]);
		free(tables);
	}
}


static void reorder(wv_pel *dst, const wv_pel* src, const int** tables, const int width, const int height, const int levels)
{
	if (tables)
	{
		int i, l, num;
		const int *idx;
		
		// output base average
		num = tables[0][0];
		idx = tables[0] + 1;
		for (i = 0; i < num; i++)
			*dst++ = src[*idx++];
		memset(dst, 0, ((width >> levels) * (height >> levels) - num) * sizeof *dst);
		dst += (width >> levels) * (height >> levels) - num;
		
		for (l = levels; l >= 1; l--)
		{
			int block_size_x = width >> l; // how big is a block ("level") of coefficients
			int block_size_y = height >> l; // in x and y?
			int pitch = 1 << (l - 1);

			// output HD coefficients 
			num = tables[levels - l][0];
			idx = tables[levels - l] + 1;
			for (i = 0; i < num; i++)
			{
				*dst++ = src[*idx + pitch]; // HD coeff in row by row order
				*dst++ = src[*idx + pitch * width]; // VD coeff in row by row order
				*dst++ = src[*idx + pitch * width + pitch]; // DD coeff in row by row order
				idx++;
			}
			memset(dst, 0, 3 * (block_size_y * block_size_x - num) * sizeof *dst);
			dst += 3 * (block_size_y * block_size_x - num);
		}
	}
	else
		memcpy(dst, src, width * height * sizeof *dst);
}


static void unreorder(wv_pel* dst, const wv_pel* src, const int** tables, const int width, const int height, const int levels)
{
	if (tables)
	{
		int i, l;
		const int *idx;

		// output base average
		idx = tables[0] + 1;
		for (i = 0; i < (width >> levels) * (height >> levels); i++)
			dst[*idx++] = *src++;
		for (l = levels; l >= 1; l--)
		{
			int block_size_x = width >> l; // how big is a block ("level") of coefficients
			int block_size_y = height >> l; // in x and y?
			int pitch = 1 << (l - 1);

			// output HD coefficients 
			idx = tables[levels - l] + 1;
			for (i = 0; i < block_size_y * block_size_x; i++)
			{
				dst[*idx + pitch] = *src++; // HD coeff in "z" order
				dst[*idx + pitch * width] = *src++; // VD coeff in in "z" order
				dst[*idx + pitch * width + pitch] = *src++; // DD coeff in in "z" order
				idx++;
			}
		}
	}
	else
		memcpy(dst, src, width * height * sizeof *dst);
}


//----------------------------------------------------------------------------
// wv_mse_to_psnr	converts mse (mean square error) into psnr (peak signal to
//				noise ratio (in dB))
//				returns 0.0 if mse <= 0.0f (instead of "infinity")
// mse			mean square error to be expressed as psnr
//----------------------------------------------------------------------------
WAVELET_DLL_API float WAVELET_DLL_CC wv_mse_to_psnr(const float mse)
{
	if (mse > 0.0f)
		return (float)(20.0 * log10(255.0 / sqrt(mse)));
	return 0.0f;
}


//----------------------------------------------------------------------------
// wv_psnr_to_mse	converts psnr (peak signal to noise ratio (in dB)) into a
//				mse (mean square error)
// psnr			psnr to be expressed as mean square error
//----------------------------------------------------------------------------
WAVELET_DLL_API float WAVELET_DLL_CC wv_psnr_to_mse(const float psnr)
{
	if (psnr > 0.0f)
		return (float)pow(255.0 / pow(10.0, psnr / 20.0), 2.0);
	return 65025.0f;
}


//----------------------------------------------------------------------------
// wv_calc_psnr	calculates the difference between the two given channels
//				and returns its psnr
// a			channel a
// b			channel b
// width		width of the rectangle to be compared
// height		height of the rectangle to be compared
// pitch		pitch of the channels in wv_pel units
//				mse (mean square error)
// pmse			address of a float where the mean-square error is written into
//				(if != NULL)
//----------------------------------------------------------------------------
WAVELET_DLL_API float WAVELET_DLL_CC wv_calc_psnr(const wv_pel *a, const wv_pel *b, const int width, const int height, const int pitch, float *pmse)
{
	double sum2 = 0.0f;
	int i, j;

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
		{
			double diff = (double)(a[i * pitch + j] - b[i * pitch + j]);

			sum2 += diff * diff;
		}

	if (sum2 > 0.0)
	{
		float mse = (float)(sum2 / (width * height)); // mean squared error
		
		if (pmse)
			*pmse = mse;
		return wv_mse_to_psnr(mse);
	}
	else
		*pmse = 0.0f;
	return 0.0f;
}


// --------------- RICE ENCODING ROUTINES ---------------

#define RICE_MIN_K 0
#define RICE_MAX_K 30
// 2 * 4 bits for num_bp and skip_bp, and 3 bits for a 1 + k + sign
#define RICE_MIN_SPACE (2 * 4 + 3)

static int rice_encode_size(const wv_pel* src, int num, int k, const int imax, int* bit_stat)
{
	int bits_written = 0;
	int cur_bp, max_bp;

	if (num == 0)
		return 0;
	
	if (imax > 0)
	{
		max_bp = cur_bp = wv_log2i(imax) - 1;
		bits_written += 4; // bit_write(num_bp, 4, f);
		if (max_bp >= 0)
		{
			int last_refinement, next_refinement;

			bits_written += 4; // bit_write(skip_bp, 4, f);
			last_refinement = 0;

			for (; cur_bp >= 0; cur_bp--)
			{
				int	mask = 1 << cur_bp;
				int ref = (2 << cur_bp) - 1;
				int num_zeros = 0;
				int idx;

				next_refinement = 0;
				for (idx = 0; idx < num; idx++)
				{
					int cur = abs(src[idx]);
					
					if (cur <= ref)
					{
						if ((cur & mask) == 0)
							num_zeros++;
						else
						{
							while (num_zeros >= (1 << k))
							{
								bits_written++; // += bit_write_single(0, f); // emit '0'
								num_zeros -= (1 << k);
								k = min(RICE_MAX_K, k + 1); // dynamically adapt k
							}
							bits_written += 1 + k + 1;
							num_zeros = 0;
							k = max(RICE_MIN_K, k - 1); // dynamically adapt k
							next_refinement++;
						}
					}
				}
				
				while (num_zeros >= (1 << k))
				{
					bits_written++; // += bit_write_single(0, f); // emit '0'
					num_zeros -= (1 << k);
					k = min(RICE_MAX_K, k + 1); // dynamically adapt k
				}
				if (num_zeros > 0)
				{
					bits_written++; // += bit_write_single(0, f); // emit '0'
					// we have indicated too many zeros, but in the decoder we know when a bitplane is done
					num_zeros = 0;
					k = max(RICE_MIN_K, k - 1); // dynamically adapt k
				}
				
				bits_written += last_refinement;
				last_refinement += next_refinement;
				
				if (bit_stat)
					bit_stat[max_bp - cur_bp] = bits_written;
			}
		}
	}
	else
	{
		bits_written += 4; // bit_write(0, 4, f); // number of bitplanes
		if (bit_stat)
			bit_stat[0] = bits_written;
	}
	
	return bits_written;
}

static int rice_encode_block(const wv_pel* src, const int num, const int skip_bp, int k, const int imax, t_bit_file* bf)
{
	int bits_written = 0;
	int max_bp, cur_bp;

	if (num == 0)
		return 0;
	
	if ((imax >> skip_bp) > 0)
	{
		max_bp = wv_log2i(imax >> skip_bp) - 1;
		bits_written += bit_write(max_bp + 1, 4, bf);
		if (max_bp >= 0)
		{
			unsigned int *refinement_cache = malloc(((num + 31) / 32) * sizeof *refinement_cache);

			bits_written += bit_write(skip_bp, 4, bf);
			for (cur_bp = wv_log2i(imax) - 1; cur_bp >= skip_bp; cur_bp--)
			{
				int	mask = 1 << cur_bp;
				int ref = (2 << cur_bp) - 1;
				int num_zeros = 0, num_refinement = 0;
				int idx, i;

				for (idx = 0; idx < num; idx++)
				{
					int cur = abs(src[idx]);

					if (cur <= ref)
					{ // not refinement
						if ((cur & mask) == 0)
							num_zeros++;
						else
						{
							i = 0;
							while (num_zeros >= (1 << k))
							{ // count the # of accumulated zeros bits
								i++;
								num_zeros -= (1 << k);
								k = min(RICE_MAX_K, k + 1); // dynamically adapt k
							}
							bits_written += bit_write(0, i, bf);
							bits_written += bit_write((1 << (k + 1)) + (num_zeros << 1) + (src[idx] < 0), 1 + k + 1, bf);
							num_zeros = 0;
							k = max(RICE_MIN_K, k - 1); // dynamically adapt k

							if (num_refinement > 0)
							{
								bits_written += num_refinement;
								for (i = 0; i < num_refinement >> 5; i++)
									bit_write(refinement_cache[i], 32, bf);
								bit_write(refinement_cache[num_refinement >> 5], num_refinement & 31, bf);
								num_refinement = 0;
							}
						}
					}
					else
					{
						refinement_cache[num_refinement >> 5] = (refinement_cache[num_refinement >> 5] << 1) + ((cur & mask) != 0);
						num_refinement++;
					}
				}
				
				i = 0;
				while (num_zeros >= (1 << k))
				{ // count the # of accumulated zeros bits
					i++;
					num_zeros -= (1 << k);
					k = min(RICE_MAX_K, k + 1); // dynamically adapt k
				}
				if (num_zeros > 0)
				{
					i++; // we have indicated too many zeros, but in the decoder we know when a bitplane is done
					k = max(RICE_MIN_K, k - 1); // dynamically adapt k
				}				
				bits_written += bit_write(0, i, bf);
				
				bits_written += num_refinement;
				for (i = 0; i < num_refinement >> 5; i++)
					bit_write(refinement_cache[i], 32, bf);
				bit_write(refinement_cache[num_refinement >> 5], num_refinement & 31, bf);
			}
			free(refinement_cache);
		}
	}
	else
		bits_written += bit_write(0, 4, bf); // number of bitplanes
	
	return bits_written;
}

static int rice_decode_block(wv_pel* dst, const int num, int k, t_bit_file* bf)
{
	if (num > 0 && dst && bf)
	{
		int cur_bp = 0;
		
		memset(dst, 0, num * sizeof *dst);

		cur_bp = bit_read(4, bf);
		if (cur_bp > 0)
		{
			int rlr_bits_left = num;
			int skip_bp;
			wv_pel sign_flag;

			skip_bp = bit_read(4, bf);
			for (cur_bp--; cur_bp >= 0; cur_bp--)
			{
				int num_zeros = 0;
				int	idx = 0;
				int refinement_left = num - rlr_bits_left;
				int cur_rlr_bits_left = rlr_bits_left;
				
				// arranged so that after all the bitplanes are read, this flag ends up in the sign-bit position (MSB)
				sign_flag = 1 << ((8 * sizeof (wv_pel)) - (1 + cur_bp));
				
				while (cur_rlr_bits_left > 0)
				{
					if (bit_read_single(bf) == 0)
					{
						num_zeros += 1 << k;
						cur_rlr_bits_left -= 1 << k;
						if (cur_rlr_bits_left < 0)
						{
							k = max(RICE_MIN_K, k - 1);
							break;
						}
						else
							k = min(RICE_MAX_K, k + 1);
					}
					else
					{
						int added_zeros;
						unsigned char encoded_sign;
						
						// read sign bit as well and increase # of 0 by 1 (2 as it's shifted >> 1), the extra zero will become the '1' bit
						added_zeros = bit_read(k + 1, bf) + 2;
						k = max(RICE_MIN_K, k - 1);
						encoded_sign = added_zeros & 1; // remember the sign
						added_zeros = min(added_zeros >> 1, cur_rlr_bits_left); // safety
						cur_rlr_bits_left -= added_zeros;
						rlr_bits_left--; // one less RLR-encoded bit
						num_zeros += added_zeros;

						for (; num_zeros > 0; idx++)
						{
							if (dst[idx] != 0)
								dst[idx] = (dst[idx] << 1) + bit_read_single(bf); // refinement
							else
								num_zeros--;
						}
						dst[idx - 1] = 0x01 + sign_flag * encoded_sign; // as we were looking for an extra zero which ended the loop
					}
				}
				// and read the last refinement bits (no need to write zeros ;))
				for (; idx < num; idx++)
					if (dst[idx] != 0)
						dst[idx] = (dst[idx] << 1) + bit_read_single(bf);
			}
			
			sign_flag--; // set all the bits below the mask
			if (skip_bp > 0)
			{ // have to dequantize as well
				int i;

				skip_bp = 1 << skip_bp;
				for (i = 0; i < num; i++)
				{
					int val = dst[i] & sign_flag;

					val = (val > 0) ? ((val + val + 1) * skip_bp) >> 1 : 0;
					if (dst[i] < 0)
						dst[i] = -val; // set sign
					else
						dst[i] = val;
				}
			}
			else
			{ // only need to "properly" set the sign
				int i;

				for (i = 0; i < num; i++)
					if (dst[i] < 0)
						dst[i] = (wv_pel)(-(dst[i] & sign_flag)); // set sign
			}
			return num;
		}
	}
	return 0;
}


// returns the next largest float
#define flarger(fl) ((fl) == 0.0f ? FLT_MIN : (fl) + (fl) * FLT_EPSILON)

//----------------------------------------------------------------------------
// wv_init_channel	initialises a channel for wavelet encoding and returns
//				a handle to it (used for finding compression parameters)
// Width		the _original_ width of the channel (also see Channel)
// Height		the _original_ height of the channel (also see Channel)
// Channel		pixel data of the channel. this _has_ to be stored with a
//				width and height that are the next-bigger power of 2 to Width
//				and Height given above. a suitable extension should be done
//				on the boundary. Unused wavelet coefficients are set to 0.
//				e.g. source channel 720x480, then Width = 720
//				and Height = 480, but the data in Channel has to be 1024x512!
// Lossless		should be set to > 0 if you only want to compress
//				losslessly (greatly speeds up computations)
//				if < 0, the error computation is approximated (but allows
//				for lossy compression [at reduced quality in some cases])
// ReorderTable	pointer to a pointer to the reorder-table. if the pointer the
//				pointers points to (huh?) is NULL, a new table is generated
//				and its address written (so we can share a single table)
// Progress		function callback used for progress display, can be NULL
// UserData		passed to the Progress function
//----------------------------------------------------------------------------
WAVELET_DLL_API t_wv_cchannel* WAVELET_DLL_CC wv_init_channel(const int Width, const int Height, const wv_pel* Channel,
	const int Lossless, int*** ReorderTable, wv_progress_function Progress, void* UserData)
{
	t_wv_cchannel* cc = NULL;

	if (Width >= 1 && Height >= 1 && Channel)
	{
		int i;
		wv_pel *channel;
		int cur, end;

		cc = malloc(sizeof *cc);

		cc->owidth = Width;
		cc->oheight = Height;
		cc->width = 1 << wv_log2i(cc->owidth - 1);
		cc->height = 1 << wv_log2i(cc->oheight - 1);
		cc->size = cc->width * cc->height;
		cc->max_levels = max(0, wv_log2i(min(cc->width, cc->height)) - 2);

		cc->reorder_table = (ReorderTable && *ReorderTable) ? *ReorderTable : init_reorder_tables(cc->width, cc->owidth, cc->height, cc->oheight, cc->max_levels, 0);
		if (ReorderTable && *ReorderTable == NULL)
			*ReorderTable = cc->reorder_table;
		
		channel = malloc(cc->size * sizeof *channel);

		cc->reordered_channel = malloc(cc->size * sizeof *cc->reordered_channel);
		memcpy(channel, Channel, cc->size * sizeof *channel); // use Channel as temp-storage
		
		wavelet_transform(channel, cc->width, cc->height, cc->max_levels);

		// reorder now sets the irrelevant wavelet coefficients to 0
		reorder(cc->reordered_channel, channel, cc->reorder_table, cc->width, cc->height, cc->max_levels);

		cc->num_blocks = cc->max_levels + 1;
		cc->size_min_block = (cc->width >> (cc->num_blocks - 1)) * (cc->height >> (cc->num_blocks - 1));

		cc->final_mse = 65536.0f;

		cc->blocks = malloc(cc->num_blocks * sizeof *cc->blocks);
		end = 0;
		for (i = 0; i < cc->num_blocks; i++)
		{ // do this before, so we know how much work we have to do
			t_wv_cblock* bl = &cc->blocks[i];
			int j;
			
			bl->offset = (i == 0) ? 0 : (cc->size_min_block << (2 * (i - 1)));
//			bl->size = (i == 0) ? cc->size_min_block : ((cc->size_min_block << (2 * i)) - bl->offset); // original
			bl->size = (i == 0) ? cc->size_min_block : 3 * cc->reorder_table[i - 1][0]; // do not write zeroes
			bl->max_value = 0;
			for (j = bl->offset; j < bl->offset + bl->size; j++)
				bl->max_value = max(bl->max_value, abs(cc->reordered_channel[j]));

			bl->max_bpp = wv_log2i(bl->max_value);
			bl->max_quant = ((Lossless > 0) || i == 0) ? 0 : bl->max_bpp;
			end += bl->max_quant + 1;
		}

		// build dictionary
		end--;
		cur = 0;
		for (i = 0; i < cc->num_blocks; i++)
		{
			t_wv_cblock* bl = &cc->blocks[i];
			int* bit_stat;
			int j;

			bit_stat = malloc(bl->max_bpp * sizeof *bit_stat); // we don't need more

			if (bl->max_bpp > 0)
				rice_encode_size(cc->reordered_channel + bl->offset, bl->size, 0, bl->max_value, bit_stat);
			
			bl->quant = malloc((bl->max_quant + 1) * sizeof *bl->quant);
			for (j = 0; j <= bl->max_quant; j++)
			{
				t_wv_quantised_block* ebl = &bl->quant[j];

				if (j == bl->max_bpp)
					ebl->compressed_size = 4;
				else
					ebl->compressed_size = bit_stat[bl->max_bpp - j - 1];
				
				ebl->mse = estimate_error((Lossless < 0) ? NULL : channel, cc->width, cc->owidth, cc->height, cc->oheight, cc->num_blocks - i, 1 << j);
				if (j > 0)
					ebl->mse = max(ebl->mse, flarger(cc->blocks[i].quant[j - 1].mse));
				
//				printf("block: %i q: %i mse: %f bits: %i\n", i, j, ebl->mse, ebl->compressed_size);
				if (Progress)
					Progress(cur, end, UserData);
				cur++;
			}
			
			free(bit_stat);
		}
		free(channel);
	}
	return cc;
}


//----------------------------------------------------------------------------
// wv_done_channel	frees a cchannel and all its associated data
// FreeReorderTable	!= 0, if you want to free the ReorderTable associated with
//					the cchannel (if a shared table is used, do this _once_!)
//----------------------------------------------------------------------------
WAVELET_DLL_API void WAVELET_DLL_CC wv_done_channel(t_wv_cchannel* CC, const int FreeReorderTable)
{
	if (CC)
	{
		int i;

		for (i = 0; i < CC->num_blocks; i++)
			free(CC->blocks[i].quant);
		free(CC->blocks);

		if (FreeReorderTable)
			free_reorder_tables(CC->reorder_table, CC->max_levels);
		free(CC->reordered_channel);

		free(CC);
	}
}


// --------------- QUANTIZER ROUTINES ---------------

static float recursive_selector_bits(t_wv_cchannel* cc, const int block_idx, const int bits_left, const float mse, unsigned char *went_down)
{
	int i, need_left;
	t_wv_cblock* bl;
	float best_lmse;
	float min_mse, max_mse;

	if (went_down)
		*went_down = 0;

	if (block_idx >= cc->num_blocks)
		return mse;

	min_mse = max_mse = mse;
	for (i = block_idx; i < cc->num_blocks; i++)
	{
		min_mse += cc->blocks[i].quant[0].mse;
		max_mse += cc->blocks[i].quant[cc->blocks[i].max_quant].mse;
	}
	
	bl = &cc->blocks[block_idx];	
	
	if (min_mse > bl->best_mse)
		return max_mse + 65536.0f; // only try this if it can possibly be better (with infite bits)
	
	best_lmse = max_mse;
	need_left = ((cc->num_blocks - 1) - block_idx) * 4; // need 4 bits for each block to indicate num_bp
	
	for (i = 0; i <= bl->max_quant; i++)
	{
		t_wv_quantised_block* ebl = &bl->quant[i];

		if (bits_left - ebl->compressed_size >= need_left)
		{ // fit
			float lmse;
			unsigned char wd;

			if (went_down)
				*went_down = 1; 
			lmse = recursive_selector_bits(cc, block_idx + 1, bits_left - ebl->compressed_size, mse + ebl->mse, &wd);
			if (lmse < bl->best_mse)
			{
				if (!wd)
				{ // we didn't fit another block, so set their quant to max
					int j;

					for (j = block_idx + 1; j < cc->num_blocks; j++)
						cc->blocks[j].best_quant = cc->blocks[j].max_quant;
				}
				bl->best_quant = i;
				bl->best_mse = best_lmse = lmse;
			}
		}
	}
	return best_lmse;
}


static int recursive_selector_mse(t_wv_cchannel* cc, const int block_idx, const int total_bits, const float mse, unsigned char *went_down)
{
	int i;
	t_wv_cblock* bl;
	int best_bits;
	float min_mse, max_mse;

	if (went_down)
		*went_down = 0;
	
	min_mse = max_mse = mse;
	for (i = block_idx; i < cc->num_blocks; i++)
	{
		min_mse += cc->blocks[i].quant[0].mse;
		max_mse += cc->blocks[i].quant[cc->blocks[i].max_quant].mse;
	}
	
	if (max_mse <= cc->max_mse)
	{
		cc->final_mse = max_mse;
		return total_bits;
	}

	if ((total_bits != 0 && cc->max_bits - total_bits < RICE_MIN_SPACE) || block_idx >= cc->num_blocks)
		return cc->max_bits + 1; // didn't meet the requirements

	if (min_mse > cc->max_mse)
		return cc->max_bits + 1; // only try this if it can possibly be better (with infite bits)
	
	best_bits = cc->max_bits + 1;
	
	bl = &cc->blocks[block_idx];	
	for (i = 0; i <= bl->max_quant; i++)
	{
		t_wv_quantised_block* ebl = &bl->quant[i];

		if (total_bits + ebl->compressed_size <= cc->max_bits)
		{ // fit
			int lbits;
			unsigned char wd;

			if (went_down)
				*went_down = 1; 
			lbits = recursive_selector_mse(cc, block_idx + 1, total_bits + ebl->compressed_size, mse + ebl->mse, &wd);
			if (lbits <= cc->max_bits && lbits < bl->best_bits)
			{
				if (!wd)
				{ // we didn't fit another block, so set their quant to max
					int j;
					
					for (j = block_idx + 1; j < cc->num_blocks; j++)
					{
						cc->blocks[j].best_quant = cc->blocks[j].max_quant;
						cc->blocks[j].best_bits = lbits;
					}
				}

				bl->best_quant = i;
				bl->best_bits = best_bits = lbits;
			}
		}
	}
	return best_bits;
}


//----------------------------------------------------------------------------
// wv_init_channel_settings	initialises compression settings for a single
//				channel, stores it to the userdefined pointer and
//				returns the number of bits used (0 on failure)
// CC			handle to the channel (gotten from wv_init_channel)
// MaxBits		(maximum) bits to be used for the channel (0 ^= quality-based)
// MaxMSE		maximum mean square error to be incurred by the compression
//				0.0f => lossless, 65536.0f => don't care
// Settings		pointer where the channel-setting handle is stored on success
//----------------------------------------------------------------------------
WAVELET_DLL_API int WAVELET_DLL_CC wv_init_channel_settings(t_wv_cchannel* CC, const int MaxBits, const float MaxMSE, t_wv_csettings** Settings)
{
	int bits = 0;

	if (CC && Settings)
	{
		int i, size;

		for (i = 0; i < CC->num_blocks; i++)
		{
			CC->blocks[i].best_bits = 1 << 30;
			CC->blocks[i].best_mse = 65536.0f;
			CC->blocks[i].best_quant = CC->blocks[i].max_quant;
		}

		CC->max_bits = size = max(0, MaxBits);
		CC->max_mse = min(65536.0f, max(0.0f, MaxMSE));

		if (CC->max_mse < 65536.0f)
		{

			if (CC->max_bits == 0)
				CC->max_bits = 1 << 30;
			size = recursive_selector_mse(CC, 0, 0, 0.0f, NULL);
		}
		else
		{
			CC->final_mse = recursive_selector_bits(CC, 0, CC->max_bits, 0.0f, NULL); // only bit-constraint
			size = 0;
			for (i = 0; i < CC->num_blocks; i++)
				size += CC->blocks[i].quant[CC->blocks[i].best_quant].compressed_size;
		}
		
		if (CC->max_bits == 0 || size <= CC->max_bits)
		{
			t_wv_csettings* set = malloc(sizeof *set);

			set->cchannel = CC;
			set->size_min_block = CC->size_min_block;
			set->emse = 0.0f;
			set->num_bits = 0;
			set->num_blocks = CC->num_blocks;
			set->blocks = malloc(CC->num_blocks * sizeof *set);
			for (i = 0; i < CC->num_blocks; i++)
			{
				set->blocks[i].max_bpp = CC->blocks[i].max_bpp;
				set->blocks[i].quant = CC->blocks[i].best_quant;
				set->blocks[i].num_bits = CC->blocks[i].quant[CC->blocks[i].best_quant].compressed_size;
				set->num_bits += set->blocks[i].num_bits;
				set->emse += CC->blocks[i].quant[CC->blocks[i].best_quant].mse;
//				printf("%i ", set->blocks[i].num_bits);
			}
//			printf("\n");
			bits = set->num_bits;
			*Settings = set;
		}
		else
			*Settings = NULL;
	}
	return bits;
}


//----------------------------------------------------------------------------
// wv_done_channel_settings	frees the channels settings
// Settings		channel-settings handle to be freed
//----------------------------------------------------------------------------
WAVELET_DLL_API void WAVELET_DLL_CC wv_done_channel_settings(t_wv_csettings* Settings)
{
	if (Settings)
	{
		free(Settings->blocks);
		free(Settings);
	}
}


//----------------------------------------------------------------------------
// wv_encode_channels	encodes a number of channels progressively
//				interleaved	into a given bit_file. it includes a small header
//				containing # channels, power of 2 width & height
//				returns the number of bits written
// NumChannels	# of channels to be encoded
// ChannelSettings	array of channel-settings handles to be used for encoding.
//				all channels / settings need to have the same width, height
//				and number of blocks to be encoded into the same file
// BF			bit-file handle to be written to (opened in WRITE-mode)
//----------------------------------------------------------------------------
WAVELET_DLL_API int WAVELET_DLL_CC wv_encode_channels(const int NumChannels, t_wv_csettings** ChannelSettings, t_bit_file* BF)
{
	int bits = 0;

	if (NumChannels > 0 && NumChannels <= wv_MAX_CHANNELS && ChannelSettings && BF)
	{
		int i;

		// check that the channels are the same size & # of blocks
		for (i = 1; i < NumChannels; i++)
		{
			if (ChannelSettings[0]->num_blocks != ChannelSettings[i]->num_blocks)
				break;
			if (ChannelSettings[0]->cchannel->width != ChannelSettings[i]->cchannel->width)
				break;
			if (ChannelSettings[0]->cchannel->height != ChannelSettings[i]->cchannel->height)
				break;
			if (ChannelSettings[0]->cchannel->size_min_block != ChannelSettings[i]->cchannel->size_min_block)
				break;
		}
		if (i == NumChannels)
		{ // they all match
			static const unsigned int magic = 0x5b765d;
			int j;

			bits += bit_write(magic, 24, BF);
			bits += bit_write(wv_CUR_HEADER_VERSION, 4, BF);
			bits += bit_write(wv_CUR_BITSTREAM_VERSION, 4, BF);

			bits += bit_write(ChannelSettings[0]->cchannel->width != ChannelSettings[0]->cchannel->owidth ||
				ChannelSettings[0]->cchannel->height != ChannelSettings[0]->cchannel->oheight, 1, BF); // write dimension flag
			if (ChannelSettings[0]->cchannel->width != ChannelSettings[0]->cchannel->owidth ||
				ChannelSettings[0]->cchannel->height != ChannelSettings[0]->cchannel->oheight)
			{
				bits += bit_write(ChannelSettings[0]->cchannel->owidth - 1, 16, BF);
				bits += bit_write(ChannelSettings[0]->cchannel->oheight - 1, 16, BF);
			}
			
			bits += bit_write(NumChannels - 1, 4, BF);
			bits += bit_write(wv_log2i(ChannelSettings[0]->cchannel->width) - 1, 4, BF);
			bits += bit_write(wv_log2i(ChannelSettings[0]->cchannel->height) - 1, 4, BF);
			bits += bit_write(ChannelSettings[0]->num_blocks - 1, 4, BF);
			bits += bit_write(wv_log2i(ChannelSettings[0]->size_min_block), 5, BF);			
			
			for (i = 0; i < ChannelSettings[0]->num_blocks; i++)
				for (j = 0; j < NumChannels; j++)
					bits += rice_encode_block(ChannelSettings[j]->cchannel->reordered_channel + ChannelSettings[j]->cchannel->blocks[i].offset, ChannelSettings[j]->cchannel->blocks[i].size, ChannelSettings[j]->blocks[i].quant, 0, ChannelSettings[j]->cchannel->blocks[i].max_value, BF); // write channel-blocks interleaved
		}
	}
	return bits;
}


//----------------------------------------------------------------------------
// wv_init_decode_channels	decodes a number of channels from a bit-file
//				written to by wv_encode_channels.
//				returns Header if header read successfully, NULL otherwise
// Header		Place to store the information (only valid if != NULL was returned)
// BF			bit-file handle to be read from (opened in READ-mode)
//----------------------------------------------------------------------------
WAVELET_DLL_API t_wv_header* WAVELET_DLL_CC wv_read_header(t_wv_header* Header, t_bit_file* BF)
{
	static const unsigned int magic = 0x5b765d;
	int header_version, bitstream_version;

	if (!Header)
		return NULL;

	Header->num_channels = 0;
	Header->width = Header->height = 0;
	Header->owidth = Header->oheight = 0;
	Header->header_version = bitstream_version = 0;

	if (!BF)
		return NULL;

	if (bit_read(24, BF) != magic)
		return NULL;

	header_version= bit_read(4, BF);
	if (header_version < wv_MIN_HEADER_VERSION || header_version > wv_CUR_HEADER_VERSION)
		return NULL;

	bitstream_version = bit_read(4, BF);
	if (bitstream_version < wv_MIN_BITSTREAM_VERSION || bitstream_version > wv_CUR_BITSTREAM_VERSION)
		return NULL;

	Header->header_version = header_version;
	Header->bitstream_version = bitstream_version;
		
	if (bit_read_single(BF))
	{ // read original dimensions
		Header->owidth = bit_read(16, BF) + 1;
		Header->oheight = bit_read(16, BF) + 1;
	}
	Header->num_channels = bit_read(4, BF) + 1;
	Header->width = 1 << bit_read(4, BF);
	Header->height = 1 << bit_read(4, BF);
	if (Header->owidth == 0)
	{ // they were equal
		Header->owidth = Header->width;
		Header->oheight = Header->height;
	}
	Header->num_blocks = bit_read(4, BF) + 1;
	Header->size_min_block = bit_read(5, BF);
	Header->size_min_block = Header->size_min_block ? (1 << (Header->size_min_block - 1)) : 0;

	if (Header->width >= 1 && Header->height >= 1 && Header->size_min_block > 0)
		return Header;
	return NULL;
}


//----------------------------------------------------------------------------
// wv_init_decode_channels	decodes a number of channels from a bit-file
//				written to by wv_encode_channels.
//				returns a structure containing all the channels (NULL on fail)
//				BF has to be aligned to the header
// Reduction	Indicates how often to halve the dimensions of the stored channels (minimum is 4x4)
// BF			bit-file handle to be read from (opened in READ-mode)
//----------------------------------------------------------------------------
WAVELET_DLL_API t_wv_dchannels* WAVELET_DLL_CC wv_init_decode_channels(int Reduction, t_bit_file* BF)
{
	t_wv_header hdr;

	if (wv_read_header(&hdr, BF) != NULL)
	{
		t_wv_dchannels* dc = NULL;
		int max_levels = max(0, wv_log2i(min(hdr.width, hdr.height)) - 2);
		int** reorder_table;
		wv_pel* tmp;
		int i, j;

		Reduction = max(0, min(Reduction, min(max_levels - 1, hdr.num_blocks - 1)));

		reorder_table = init_reorder_tables(hdr.width, hdr.owidth, hdr.height, hdr.oheight, max_levels - Reduction, Reduction);
		
		hdr.width >>= Reduction;
		hdr.height >>= Reduction;
		hdr.owidth >>= Reduction;
		hdr.oheight >>= Reduction;
		hdr.num_blocks -= Reduction;
		max_levels -= Reduction;

		dc = malloc(sizeof *dc);
		dc->channels = malloc(hdr.num_channels * sizeof *dc->channels);
		for (i = 0; i < hdr.num_channels; i++)
			dc->channels[i] = malloc(hdr.width * hdr.height * sizeof *dc->channels[i]);
		
		for (i = 0; i < hdr.num_blocks; i++)
		{
			int ofs = (i == 0) ? 0 : (hdr.size_min_block << (2 * (i - 1)));
//			int size = (i == 0) ? size_min_block : ((size_min_block << (2 * i)) - ofs); // original
			int size = (i == 0) ? hdr.size_min_block : 3 * reorder_table[i - 1][0]; // do not read zeroes

			for (j = 0; j < hdr.num_channels; j++)
				rice_decode_block(dc->channels[j] + ofs, size, 0, BF);
		}
		
		tmp = malloc(hdr.width * hdr.height * sizeof *tmp);

		for (i = 0; i < hdr.num_channels; i++)
		{
			unreorder(tmp, dc->channels[i], reorder_table, hdr.width, hdr.height, max_levels);
			inverse_wavelet_transform(tmp, hdr.width, hdr.height, max_levels);
			memcpy(dc->channels[i], tmp, hdr.width * hdr.height * sizeof *dc->channels[i]);
		}

		free(tmp);
		free_reorder_tables(reorder_table, max_levels);
		
		dc->hdr = hdr;
		return dc;
	}
	return NULL;
}


//----------------------------------------------------------------------------
// wv_done_decode_channels	frees the structure allocated
//				by wv_init_decode_channels.
// dc			decompressed channel structure to be freed
//----------------------------------------------------------------------------
WAVELET_DLL_API void WAVELET_DLL_CC wv_done_decode_channels(t_wv_dchannels* dc)
{
	if (dc)
	{
		int i;

		for (i = 0; i < dc->hdr.num_channels; i++)
			free(dc->channels[i]);
		free(dc->channels);
		free(dc);
	}
}