/* -*- Mode: C; tab-width: 4; -*- */
/* npwebwko.c
 * WebWKO - a Win32 browser plug-in that displays WKO image files
 * By Daniel Vollmer, based on WebBMP by Jason Summers
 */
#define PRGNAME "WebWKO Plug-in"
#define VERS    "1.0.0"

/* Note: Depending on your compiler, npwin.cpp may need to be altered by
 * wrapping the g_pNavigatorFuncs definition with 'extern "C"' as such:
 *
 *  extern "C" {
 *  NPNetscapeFuncs* g_pNavigatorFuncs = 0;
 *  }
 *
 */

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <process.h>
#include <time.h>
#include "npupp.h"
#include "npapi.h"
#include "resource.h"

#include "../wavelet/bit.h"
#include "../wavelet/wavelet.h"
#include "../wavelet/mchannel.h"

#define ST_NOTLOADED	0
#define ST_LOADED	1

#define DELAY_PER_3K 2
#define INITIAL_UPDATE_BITS 1024

#define INVALID_THREAD (NULL)

/* plugin instance-specific data */
typedef struct plugininstance_struct
{
	NPWindow*  fWindow;
	uint16     fMode;
	HWND       fhWnd;
	WNDPROC    fDefaultWindowProc;
	NPP        instance;

	char url[2048];  /* the url of the stream */
	int status;      /* is the image finished loading? (either ST_LOADED or ST_NOTLOADED) */
	int streamend;   /* size of stream, or 0 if unknown */
	int valid;  /* are the height, width, etc. valid? */
	unsigned int last_bits_drawn;
	time_t last_time_drawn;
	unsigned int update_bits;
	
	BITMAPINFO bi;
	HBITMAP dib;
	unsigned int *dib_bits;

	int	wvh_is_yuv;
	t_wv_header wvh;
	
	CRITICAL_SECTION bitmap_lock;
	CRITICAL_SECTION lpbits_lock;
	HANDLE decode_thread, last_thread;

	unsigned char *lpbits;    /* direct pointer to the image bits */
	unsigned int bits_read;	/* number of _bits_ stored in lpbits */
	int bytes_alloc;    /* number of _bytes_ allocated for lpbits */
} PluginInstance;


/* global variables */

extern NPNetscapeFuncs* g_pNavigatorFuncs;

const char* gInstanceLookupString = "pdata";
HMODULE hInst;
HMENU hmenuContext=NULL;

/* forward declarations of functions */
LRESULT CALLBACK DlgProcAbout(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK PluginWindowProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam);
void UpdateBitmap(PluginInstance *This);
void PaintBitmap(PluginInstance *This, HDC hdc, RECT *rect);
void ContextMenu(PluginInstance *This, int x, int y, HWND hwnd);


/* ----------------------------------------------------------------------- */

BOOL APIENTRY DllMain(HANDLE hModule, DWORD reason, LPVOID lpReserved)
{
    switch(reason) {
    case DLL_PROCESS_ATTACH:
		hInst=hModule;
		break;
    case DLL_THREAD_ATTACH:
		break;
    case DLL_THREAD_DETACH:
		break;
    case DLL_PROCESS_DETACH:
		break;
    }
    return TRUE;
}


void warn(PluginInstance *This, char *fmt, ...)
{
	va_list ap;
	char buf[2048];
	HWND hwnd;

	va_start(ap, fmt);
	vsprintf(buf,fmt, ap);
	va_end(ap);

	if (This)
		hwnd=This->fhWnd;
	else
		hwnd=NULL;

	MessageBox(hwnd,buf,"WebWKO Plug-in",MB_OK|MB_ICONWARNING);
}


/* if enough of the image is loaded, extract its width, height, etc. */
void GetImageInfo(PluginInstance *This)
{
	t_bit_file *bf;

	/*********
	1 bit:	is yuv
	24 bits:magic number
	4 bits:	header version
	4 bits: bitstream version
	1 bit:	have non-pow2 dimensions
		16 bit:	width - 1
		16 bit: height - 1
	4 bits:	num channels - 1
	4 bits:	pow2 width
	4 bits:	pow2 height
	4 bits:	num blocks - 1
	5 bits:	size min block
	-----------------
	55 bits (dimension bit 0) or 87 bits (dimension bit 1)
	**********/
	
	if (!This->lpbits || This->bits_read < 87)
		return;

	bf = bit_open(This->lpbits, "rm", This->bits_read);
	if (bf)
	{
		t_wv_header hdr;

		This->wvh_is_yuv = bit_read(1, bf);
		if (wv_read_header(&hdr, bf) != NULL)
		{
			This->wvh = hdr;
			if (This->wvh.width >= 1 && This->wvh.height >= 1 && This->wvh.size_min_block > 0)
			{
				This->bi.bmiHeader.biSize = sizeof This->bi.bmiHeader;
				This->bi.bmiHeader.biWidth = This->wvh.owidth;
				This->bi.bmiHeader.biHeight = This->wvh.oheight;
				This->bi.bmiHeader.biPlanes = 1;
				This->bi.bmiHeader.biBitCount = 32;
				This->bi.bmiHeader.biCompression = BI_RGB;
				This->bi.bmiHeader.biSizeImage = This->wvh.owidth * This->wvh.oheight * 4;
				This->bi.bmiHeader.biXPelsPerMeter = This->bi.bmiHeader.biYPelsPerMeter = 0;
				This->bi.bmiHeader.biClrUsed = This->bi.bmiHeader.biClrImportant = 0;
				
				This->valid = 1;
			}
		}
		bit_close(bf, NULL);
	}
}


/* ----------------------------------------------------------------------- */

/* Global initialization */
NPError NPP_Initialize(void)
{
	hmenuContext = LoadMenu(hInst,"CONTEXTMENU");
	return NPERR_NO_ERROR;
}

/* Global shutdown */
void NPP_Shutdown(void)
{
	if (hmenuContext)
		DestroyMenu(hmenuContext);
}

/* Once-per-instance initialization */
NPError NPP_New(NPMIMEType pluginType,NPP instance,uint16 mode,
		int16 argc,char* argn[],char* argv[],NPSavedData* saved)
{
	PluginInstance* This;
	
	if (!instance)
		return NPERR_INVALID_INSTANCE_ERROR;
	
	instance->pdata = malloc(sizeof (PluginInstance));
	This = (PluginInstance*)instance->pdata;
	if (!This)
	    return NPERR_OUT_OF_MEMORY_ERROR;

	memset(This, 0, sizeof *This);
	/* record some info for later lookup */
	This->fWindow = NULL;
	This->fMode = mode;
	
	This->fhWnd = NULL;
	This->fDefaultWindowProc = NULL;
	This->instance = instance;  /* save the instance id for reverse lookups */

	This->lpbits = NULL;
	This->bits_read = 0;
	This->bytes_alloc=0;
	This->status=ST_NOTLOADED;
	This->valid=0;

	This->dib = NULL;
	This->dib_bits = NULL;

	This->decode_thread = This->last_thread = INVALID_THREAD;
	InitializeCriticalSection(&This->bitmap_lock);
	InitializeCriticalSection(&This->lpbits_lock);

	return NPERR_NO_ERROR;
}

NPError NPP_Destroy(NPP instance, NPSavedData** save)
{
	PluginInstance* This;

	if (!instance)
		return NPERR_INVALID_INSTANCE_ERROR;
	This = (PluginInstance*)instance->pdata;
	if (!This)
		return NPERR_INVALID_INSTANCE_ERROR;

	if (This->fhWnd)
	{ // If we have a window, clean it up (un-subclass it)
		SetWindowLong( This->fhWnd, GWL_WNDPROC, (LONG)This->fDefaultWindowProc);
		This->fDefaultWindowProc = NULL;
		This->fhWnd = NULL;
	}

	if (This->last_thread != INVALID_THREAD)
	{
		WaitForSingleObject(This->last_thread, INFINITE);
		CloseHandle(This->last_thread);
		This->last_thread = INVALID_THREAD;
	}

	DeleteCriticalSection(&This->bitmap_lock);
	DeleteCriticalSection(&This->lpbits_lock);
	
	if (This->lpbits)
	{
		free(This->lpbits);
		This->lpbits = NULL;
		This->bytes_alloc = 0;
		This->bits_read = 0;
	}

	if (This->dib)
	{
		DeleteObject(This->dib);
		This->dib = NULL;
	}

	if (This)
	{
		if (instance->pdata)
		{
			free(instance->pdata);
			instance->pdata = NULL;
		}
	}
	return NPERR_NO_ERROR;
}

jref NPP_GetJavaClass(void)
{
	return NULL;
}

/* Browser is providing us with a window */
NPError NPP_SetWindow(NPP instance, NPWindow* window)
{
	NPError result = NPERR_NO_ERROR;
	PluginInstance* This;

	if (!instance)
		return NPERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*)instance->pdata;

	/* if( This->fWindow != NULL )  -- from example in Netscape's API, but not really correct */

	if (This->fhWnd)
	{   /* If we already have a window... */
		if ((!window) || (!window->window))
		{
			/* There is now no window to use. get rid of the old
			 * one and exit. */
			SetWindowLong(This->fhWnd, GWL_WNDPROC, (LONG)This->fDefaultWindowProc);
			This->fDefaultWindowProc = NULL;
			This->fhWnd = NULL;
			This->fWindow = window;
			return NPERR_NO_ERROR;
		}
		else if (This->fhWnd == (HWND)window->window)
		{
			/* The new window is the same as the old one. Redraw and get out. */
			This->fWindow = window;

			InvalidateRect(This->fhWnd, NULL, FALSE);
			return NPERR_NO_ERROR;
		}
		else
		{
			/* Unsubclass the old window, so that we can subclass the new
			 * one later. */
			SetWindowLong(This->fhWnd, GWL_WNDPROC, (LONG)This->fDefaultWindowProc);
			This->fDefaultWindowProc = NULL;
			This->fhWnd = NULL;
		}
	}
	else if ((!window) || (!window->window))
	{
		/* We can just get out of here if there is no current
		 * window and there is no new window to use. */
		This->fWindow=window;

		return NPERR_NO_ERROR;
	}

	/* Subclass the new window so that we can begin drawing and
	 * receiving window messages. */
	This->fDefaultWindowProc = (WNDPROC)SetWindowLong((HWND)window->window, GWL_WNDPROC, (LONG)PluginWindowProc);
	This->fhWnd = (HWND)window->window;
	SetProp(This->fhWnd, gInstanceLookupString, (HANDLE)This);

	This->fWindow = window;

	InvalidateRect(This->fhWnd, NULL, FALSE);
//	UpdateWindow(This->fhWnd);

	return result;
}

/*+++++++++++++++++++++++++++++++++++++++++++++++++
 * NPP_NewStream:
 * Notifies an instance of a new data stream and returns an error value. 
 * 
 * NPP_NewStream notifies the instance denoted by instance of the creation of
 * a new stream specifed by stream. The NPStream* pointer is valid until the
 * stream is destroyed. The MIME type of the stream is provided by the
 * parameter type. 
 +++++++++++++++++++++++++++++++++++++++++++++++++*/
NPError  NPP_NewStream(NPP instance,NPMIMEType type,NPStream *stream, 
	      NPBool seekable,uint16 *stype) {
	PluginInstance* This;

	if(instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*)instance->pdata;
	if(!This)
		return NPERR_GENERIC_ERROR;

	/* save the URL for later */
	strncpy(This->url, stream->url, 2048);
	This->url[2047]='\0';

	*stype = NP_NORMAL;

	if (This->last_thread != INVALID_THREAD)
	{
		WaitForSingleObject(This->last_thread, INFINITE);
		CloseHandle(This->last_thread);
		This->last_thread = INVALID_THREAD;
	}

	if(This->lpbits)
	{
		free(This->lpbits);
		This->lpbits = NULL;
		This->bytes_alloc=0;
	}
	if (This->dib)
	{
		DeleteObject(This->dib);
		This->dib = NULL;
		This->dib_bits = NULL;
	}
	This->status=ST_NOTLOADED;
	This->streamend = stream->end;
	This->bits_read = 0;
	This->last_bits_drawn = 0;
	This->last_time_drawn = 0;
	This->last_time_drawn = time(NULL);
	This->update_bits = INITIAL_UPDATE_BITS;
	This->valid = 0;
	return NPERR_NO_ERROR;
}

int32 NPP_WriteReady(NPP instance, NPStream *stream)
{
	/* Number of bytes ready to accept in NPP_Write() */
	/* We can handle any amount, so just return some really big number. */
	return (int32)0x0FFFFFFF;
//	return (int32)INITIAL_UPDATE_BITS / 8;
}

/*+++++++++++++++++++++++++++++++++++++++++++++++++
 * NPP_Write:
 * Delivers data from a stream and returns the number of bytes written. 
 * 
 * NPP_Write is called after a call to NPP_NewStream in which the plug-in
 * requested a normal-mode stream, in which the data in the stream is delivered
 * progressively over a series of calls to NPP_WriteReady and NPP_Write. The
 * function delivers a buffer buf of len bytes of data from the stream identified
 * by stream to the instance. The parameter offset is the logical position of
 * buf from the beginning of the data in the stream. 
 * 
 * The function returns the number of bytes written (consumed by the instance).
 * A negative return value causes an error on the stream, which will
 * subsequently be destroyed via a call to NPP_DestroyStream. 
 * 
 * Note that a plug-in must consume at least as many bytes as it indicated in the
 * preceeding NPP_WriteReady call. All data consumed must be either processed
 * immediately or copied to memory allocated by the plug-in: the buf parameter
 * is not persistent. 
 +++++++++++++++++++++++++++++++++++++++++++++++++*/
int32 NPP_Write(NPP instance, NPStream *stream, int32 offset, int32 len, void *buffer)
{
	PluginInstance* This;
	int32 new_alloc;

	if (!instance)
		return -1;
	This = (PluginInstance*)instance->pdata;
	if (!This)
		return -1;

	new_alloc = This->bytes_alloc;
	if (!This->lpbits)
		new_alloc = This->streamend > 0 ? This->streamend : offset + len;
	new_alloc = max(new_alloc, offset + len);
	if (new_alloc != This->bytes_alloc)
	{
		EnterCriticalSection(&This->lpbits_lock);
		This->lpbits = realloc(This->lpbits, new_alloc);
		This->bytes_alloc = new_alloc;
		This->bits_read = (offset + len) * 8; // hopefully we read everything _before_ that, too...
		LeaveCriticalSection(&This->lpbits_lock);
		if (!This->lpbits)
		{
			warn(This,"Out of memory");
			return -1;
		}
		memset(This->lpbits + This->bytes_alloc, 0, new_alloc - This->bytes_alloc);
	}
	else
		This->bits_read = (offset + len) * 8; // hopefully we read everything _before_ that, too...
	
	memcpy(This->lpbits + offset, buffer, len);
	
	if (!This->valid)
		GetImageInfo(This);
	if (This->fhWnd && This->valid && This->bits_read - This->last_bits_drawn >= This->update_bits)
	{ /* progressive display */
		time_t ct;
		int timeofs;

		/* Since we repaint the whole image every time we do a progressive
		 * display update, it can slow things down if we do this too often,
		 * especially if the image is being scaled to a different size. To
		 * prevent this, we make sure a certain amount of time has elapsed
		 * since the last update */
		ct = time(NULL);

		timeofs = ((This->wvh.width + This->wvh.height) * This->wvh.num_channels * DELAY_PER_3K) / (3 * 1024);
		if (ct - This->last_time_drawn >= timeofs)
			UpdateBitmap(This);
	}

	return len; /* The number of bytes accepted -- we always accept them all. */
}



/* DestroyStream gets called after the file has finished loading,
 */
NPError NPP_DestroyStream(NPP instance, NPStream *stream, NPError reason)
{
	PluginInstance* This;

	if (!instance)
		return NPERR_INVALID_INSTANCE_ERROR;
	This = (PluginInstance*)instance->pdata;
	if (!This)
		return NPERR_INVALID_INSTANCE_ERROR;

	if (!This->valid)
		GetImageInfo(This);
	
	if (reason == NPRES_DONE)
	{
		This->status = This->valid ? ST_LOADED : ST_NOTLOADED;

		if (!This->valid)
			warn(This, "Invalid WKO image");
	}
	
	if (This->fhWnd && This->bits_read - This->last_bits_drawn > 0)
	{
		if (This->last_thread != INVALID_THREAD)
		{
			WaitForSingleObject(This->last_thread, INFINITE);
			CloseHandle(This->last_thread);
			This->last_thread = INVALID_THREAD;
		}
/*		OutputDebugString("Final Update: ");
		OutputDebugString(This->url);
		OutputDebugString("\n");*/
		UpdateBitmap(This); // we also update the window again if we cancel loading
	}
	
	return NPERR_NO_ERROR;
}

/* This should not get called, since we aren't using the NP_ASFILE mode */
void NPP_StreamAsFile(NPP instance, NPStream *stream, const char* fname)
{
	return;
}

/* Print embedded plug-in (via the browser's Print command) */
void NPP_Print(NPP instance, NPPrint* printInfo)
{
	PluginInstance* This;
	
	if (!instance)
		return;
	This = (PluginInstance*) instance->pdata;
	if (!This)
		return;

	if (!This->valid)
		return;

	if(printInfo == NULL) {
		/* Though it's not documented, if a Netscape user tries to print a full-page
		 * plugin, the browser will call NPP_Print with printInfo set to NULL.
		 * At this point you should print your plug-in in whatever manner you
		 * choose. */

		/* PrintFullPage();  */     /* not implemented */
		return;
	}
	
	if (printInfo->mode == NP_FULL) {
		/* the plugin is full-page, and the browser is giving it a chance
		 * to print in the manner of its choosing */
		void* platformPrint = printInfo->print.fullPrint.platformPrint;
		NPBool printOne =  printInfo->print.fullPrint.printOne;
		
		/* Setting this to FALSE and returning *should* cause the browser to
		 * call NPP_Print again, this time with mode=NP_EMBED.
		 * However, that doesn't happen with any browser I've ever seen :-(.
		 * Instead of the following line, you will probably need to implement
		 * printing yourself.  You also might as well set pluginPrinted to TRUE,
		 * though every browser I've tested ignores it. */
		printInfo->print.fullPrint.pluginPrinted = FALSE;
		/*  or */
		/*  PrintFullPage();
		 *  printInfo->print.fullPrint.pluginPrinted = TRUE;  */
	}
	else {	/* If not fullscreen, we must be embedded */
		HDC pdc;
		int prevstretchmode;
		NPWindow* printWindow;
		
		if (This->status != ST_LOADED)
			return;

		printWindow= &(printInfo->print.embedPrint.window);

		/* embedPrint.platformPrint is a Windows device context in disguise */

		/* The definition of NPWindow changed between API verion 0.9 and 0.11,
		 * increasing in size from 28 to 32 bytes. This normally makes it
		 * impossible for version 0.9 browsers to print version 0.11 plugins
		 * (because the platformPrint field ends up at the wrong offset) --
		 * unless the plugin takes special care to detect this situation.
		 * To work around it, if we are compiled with API 0.11 or higher,
		 * and the browser is version 0.9 or earlier, we look for the HDC
		 * 4 bytes earlier, at offset 28 instead of 32 (of the embedPrint
		 * sub-structure).
		 */

		if(sizeof(NPWindow)>28 &&     /* i.e. is plugin API >= 0.11? */
			     HIBYTE(g_pNavigatorFuncs->version)==0 &&
		         LOBYTE(g_pNavigatorFuncs->version)<=9) {
			char *tmpc;
			HDC  *tmph;

			tmpc= (char*)&(printInfo->print.embedPrint);
			tmph= (HDC*)&tmpc[28];
			pdc=  *tmph;  /* I know this could be all be done in 1 line, */
			              /* it would look too messy for my taste. */
		}
		else {
			pdc= (HDC) (printInfo->print.embedPrint.platformPrint);
		}

		if (!This->dib)
			return;

		prevstretchmode = SetStretchBltMode(pdc, HALFTONE);
		StretchDIBits(pdc,
			printWindow->x, printWindow->y, printWindow->width, printWindow->height,
			0, 0, This->wvh.owidth, This->wvh.oheight,
			This->dib_bits, &This->bi, DIB_RGB_COLORS, SRCCOPY);
		if (prevstretchmode)
			SetStretchBltMode(pdc, prevstretchmode);
	}
}

void NPP_URLNotify(NPP instance, const char* url, NPReason reason, void* notifyData)
{
	return;
}

int16 NPP_HandleEvent(NPP instance, void* event)
{
	return 0;
}

/**********************************************************************/

DWORD WINAPI DecodeFunc(PluginInstance *This)
{
/*	OutputDebugString("Thread Enter: ");
	OutputDebugString(This->url);
	OutputDebugString("\n");
*/	if (This && This->valid && This->bits_read - This->last_bits_drawn > 0 && This->dib)
	{
		t_bit_file *bf;

/*		OutputDebugString("Thread Working: ");
		OutputDebugString(This->url);
		OutputDebugString("\n");
*/		This->last_time_drawn = time(NULL);
		This->update_bits *= 3;
		EnterCriticalSection(&This->lpbits_lock);
		This->last_bits_drawn = This->bits_read;
		bf = bit_open(This->lpbits, "rm", This->bits_read);	
		if (bf)
		{
			t_wv_dchannels* dc;
			int ycbcr;

			ycbcr = bit_read(1, bf);
			dc = wv_init_decode_channels(0, bf);
			bit_close(bf, NULL);
			LeaveCriticalSection(&This->lpbits_lock);

			if (dc)
			{
				wv_pel *r, *g, *b;
				int y, x;

				if (ycbcr && dc->hdr.num_channels >= 3)
				{ // ycbcr -> rgb
					for (y = 0; y < dc->hdr.oheight; y++)
						wv_ycbcr_to_rgb(dc->hdr.width, dc->channels[0] + y * dc->hdr.width, dc->channels[1] + y * dc->hdr.width,
							dc->channels[2] + y * dc->hdr.width);
				}

				r = dc->channels[0];
				g = dc->hdr.num_channels > 1 ? dc->channels[1] : dc->channels[dc->hdr.num_channels - 1];
				b = dc->hdr.num_channels > 2 ? dc->channels[2] : dc->channels[dc->hdr.num_channels - 1];
				EnterCriticalSection(&This->bitmap_lock);
				GdiFlush();
				for (y = 0; y < This->wvh.oheight; y++)
				{
					unsigned int ofs = y * This->wvh.width;
					unsigned int* dst = This->dib_bits + (This->bi.bmiHeader.biHeight - y - 1) * This->bi.bmiHeader.biWidth;

					for (x = 0; x < This->wvh.owidth; x++)
					{
						*dst++ = RGB(min(255, max(0, b[ofs])), min(255, max(0, g[ofs])), min(255, max(0, r[ofs])));
						ofs++;
					}
				}
				LeaveCriticalSection(&This->bitmap_lock);
				wv_done_decode_channels(dc);
			}
		}
		else
			LeaveCriticalSection(&This->lpbits_lock);

		InvalidateRect(This->fhWnd, NULL, FALSE);
//		UpdateWindow(This->fhWnd);
/*		OutputDebugString("Thread Done: ");
		OutputDebugString(This->url);
		OutputDebugString("\n");
*/	}
	This->decode_thread = INVALID_THREAD;
	return 0;
} 

void UpdateBitmap(PluginInstance *This)
{
	if (This && This->valid && This->decode_thread == INVALID_THREAD && (This->bits_read - This->last_bits_drawn > 0 || !This->dib))
	{
		if (!This->dib)
			This->dib = CreateDIBSection(NULL, &This->bi, DIB_RGB_COLORS, &This->dib_bits, NULL, 0);
		if (This->dib && This->dib_bits)
		{
			DWORD tid;

			if (This->last_thread != INVALID_THREAD)
			{
				WaitForSingleObject(This->last_thread, INFINITE);
				CloseHandle(This->last_thread);
				This->last_thread = INVALID_THREAD;
			}
			This->decode_thread = This->last_thread = CreateThread(NULL, 0, DecodeFunc, This, 0, &tid);
//			This->decode_thread = _beginthread(DecodeFunc, 0, This);
//			DecodeFunc(This);
		}
	}
}

/* draw bitmap from DIB 
 * rect: I am abusing the definition of RECT, and using rect->right,bottom as
 *       width,height instead of absolute coordinates
 */
void PaintBitmap(PluginInstance *This, HDC hdc, RECT *rect1) 
{
	int prevstretchmode;
	RECT rect;

	/* We consider displaying it only if it is completely loaded, or if
	 * the complete header has been loaded */
	if (!This->valid || !This->dib)
		return;
	
	/* make a local copy */
	rect.left = rect1->left;    rect.top=rect1->top;
	rect.right = rect1->right;  rect.bottom=rect1->bottom;

	/* Show full-screen images their natural size, unless they don't fit
	 * in the browser, in which case scale them down (the Right Thing to do
	 * would be to create scroll bars, etc.) */
	if (This->fMode == NP_FULL)
	{
		int w,h,x,y;
		HPEN oldpen;
		HBRUSH oldbrush;

		w = This->wvh.owidth;
		h = This->wvh.oheight;

		/* reduce size if needed */
		if (w > rect.right)
		{
			h = (int)(((double)h * (double)rect.right)/(double)w);
			w = rect.right;
		}
		if (h > rect.bottom)
		{
			w = (int)(((double)w * (double)rect.bottom)/(double)h);
			h = rect.bottom;
		}

		/* center the image if there's room */
		if (w < rect.right) 
			x = (rect.right - w) / 2;
		else
			x = 0;
		if (h < rect.bottom)
			y = (rect.bottom - h) / 2;
		else
			y = 0;

		rect.left = x;
		rect.right = w;
		rect.top = y;
		rect.bottom = h;

		oldpen = SelectObject(hdc, GetStockObject(NULL_PEN));
		oldbrush = SelectObject(hdc, GetStockObject(WHITE_BRUSH));

		Rectangle(hdc,rect1->left,rect1->top, rect1->left+rect1->right+1,rect1->top+rect1->bottom+1);

		if (oldpen)
			SelectObject(hdc, oldpen);
		if (oldbrush)
			SelectObject(hdc, oldbrush);
	}

	prevstretchmode = SetStretchBltMode(hdc, HALFTONE);
	EnterCriticalSection(&This->bitmap_lock);
	StretchDIBits(hdc,
		rect.left, rect.top, rect.right, rect.bottom,
		0, 0, This->wvh.owidth, This->wvh.oheight,
		This->dib_bits, &This->bi, DIB_RGB_COLORS, SRCCOPY);
	LeaveCriticalSection(&This->bitmap_lock);
	if (prevstretchmode)
		SetStretchBltMode(hdc, prevstretchmode);
}

void CopyToClipboard(PluginInstance *This, unsigned char *mem, int size, UINT format)
{
	HGLOBAL hClip;
	LPVOID lpClip;

	if (!mem)
		return;

	if (!OpenClipboard(NULL))
	{
		warn(This,"Can't open the clipboard");
		return;
	}

	if (EmptyClipboard())
	{
		hClip = GlobalAlloc(GMEM_ZEROINIT|GMEM_MOVEABLE|GMEM_DDESHARE,size);
		lpClip = GlobalLock(hClip);
		if (lpClip)
		{
			CopyMemory(lpClip, mem, size);
			GlobalUnlock(hClip);
			if(!SetClipboardData(format,hClip))
				warn(This, "Can't set clipboard data");
		}
		else
			warn(This, "Can't allocate memory for clipboard");
	}
	else
		warn(This, "Can't clear the clipboard");
	CloseClipboard();
}


int SaveToBMPFile(PluginInstance *This, char *filename)
{
	HANDLE hFile;
	BITMAPFILEHEADER hdr;
	DWORD written;

	if (!This->dib)
		return 0;

	hFile = CreateFile(filename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		warn(This, "Could not save file '%s'", filename);
		return 0;
	}

	hdr.bfType = 0x4d42; // BM
    hdr.bfSize = (DWORD) (sizeof(BITMAPFILEHEADER) + This->bi.bmiHeader.biSize + This->bi.bmiHeader.biSizeImage); 
    hdr.bfReserved1 = 0; 
    hdr.bfReserved2 = 0; 
    // Compute the offset to the array of color indices. 
    hdr.bfOffBits = (DWORD) sizeof(BITMAPFILEHEADER) + This->bi.bmiHeader.biSize; 
	WriteFile(hFile, (LPVOID)&hdr, sizeof hdr, &written, NULL);
	WriteFile(hFile, (LPVOID)&This->bi.bmiHeader, This->bi.bmiHeader.biSize, &written, NULL);
	WriteFile(hFile, (LPVOID)This->dib_bits, This->bi.bmiHeader.biSizeImage, &written, NULL);

	CloseHandle(hFile);
	return 1;
}

/* Try to make a filename from the url. Caller must provide fn[256] buffer.
 * This function attempts to extract a bitmap filename from a URL,
 * but if it doesn't look like it contains an appropriate name,
 * it leaves it blank. */
void url2filename(char *fn, char *url)
{
	int title,ext,i;

	strcpy(fn,"");
	ext=0;   /* position of the file extention */
	title=0; /* position of the base filename */
	for(i=0;url[i];i++) {
		if(url[i]=='.') ext=i+1;
		if(url[i]=='/') title=i+1;
		if(url[i]==':') title=i+1;
		if(url[i]=='=') title=i+1;
	}

	if(!_stricmp(&url[ext],"wko")) {
		strncpy(fn,&url[title],MAX_PATH);
		fn[MAX_PATH-1]='\0';
	}
}

/* Write the image to a local file; see SaveToBMPFile() */
void SaveImage(PluginInstance *This)
{
	OPENFILENAME ofn;
	char fn[MAX_PATH];

	if(This->status != ST_LOADED) {
		warn(This,"Image not loaded -- can't save");
		return;
	}

	if(strlen(This->url)) {
		url2filename(fn,This->url);
	}
	else {
		strcpy(fn,"");
	}

	ZeroMemory(&ofn,sizeof(OPENFILENAME));
	ofn.lStructSize=sizeof(OPENFILENAME);
	ofn.hwndOwner=This->fhWnd;
	ofn.lpstrFilter="Wavelet Kompressed (*.wko)\0*.wko\0Bitmaps (*.bmp)\0*.bmp\0\0";
	ofn.nFilterIndex=1;
	ofn.lpstrTitle="Save Image As...";
	ofn.lpstrFile=fn;
	ofn.nMaxFile=MAX_PATH;
	ofn.Flags=OFN_PATHMUSTEXIST|OFN_HIDEREADONLY|OFN_OVERWRITEPROMPT;
	ofn.lpstrDefExt="wko";

	if(GetSaveFileName(&ofn))
	{
		if (strcmpi(&ofn.lpstrFile[ofn.nFileExtension], "bmp") == 0)
			SaveToBMPFile(This,ofn.lpstrFile);
		else
		{
			FILE *f = fopen(ofn.lpstrFile, "wb");

			if (f)
			{
				fwrite(This->lpbits, This->bits_read / 8, 1, f);
				fclose(f);
			}
			else
				warn(This, "Could not save file '%s'", ofn.lpstrFile);
		}
	}
}

void AboutDialog(PluginInstance *This)
{
	DialogBoxParam(hInst,"ABOUTDLG",This->fhWnd,(DLGPROC)DlgProcAbout,(LPARAM)This);
}

void ContextMenu(PluginInstance *This, int x, int y, HWND hwnd)
{
	HMENU sub;
	int cmd;
	POINT pt;

	pt.x=pt.y=0;
	GetCursorPos(&pt);
	x=pt.x; y=pt.y;

	sub=GetSubMenu(hmenuContext,0);

	cmd=TrackPopupMenuEx(sub, TPM_LEFTALIGN|TPM_TOPALIGN|TPM_NONOTIFY|TPM_RETURNCMD|
		TPM_RIGHTBUTTON,x,y,hwnd,NULL);

	switch(cmd)
	{
		case ID_SAVEAS:
			SaveImage(This);
		break;
		case ID_COPYIMAGE:
			if (This->status==ST_LOADED && This->dib)
			{
				HBITMAP bmp;

				bmp = CreateBitmap(This->bi.bmiHeader.biWidth, abs(This->bi.bmiHeader.biHeight), 1, 32, This->dib_bits); 
				
				if (!OpenClipboard(NULL))
				{
					warn(This,"Can't open the clipboard");
					return;
				}

				if (EmptyClipboard())
				{
					if(!SetClipboardData(CF_BITMAP, bmp))
						warn(This, "Can't set clipboard data");
				}
				else
					warn(This, "Can't clear the clipboard");
				CloseClipboard();
				DeleteObject(bmp);
			}
			else
				warn(This,"No image to copy");
		break;
		case ID_COPYURL:
			if (strlen(This->url))
				CopyToClipboard(This,This->url,strlen(This->url)+1,CF_TEXT);
			else
				warn(This,"No link to copy");
		break;
		case ID_VIEWIMAGE:
			if (strlen(This->url)) 
				NPN_GetURL(This->instance,This->url, "_self");
			break;
		case ID_ABOUT:
			AboutDialog(This);
		break;
	}
}

/*+++++++++++++++++++++++++++++++++++++++++++++++++
 * PluginWindowProc
 * Handle the Windows window-event loop.
 +++++++++++++++++++++++++++++++++++++++++++++++++*/
LRESULT CALLBACK PluginWindowProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	PluginInstance* This = (PluginInstance*) GetProp(hWnd, gInstanceLookupString);
	if(!This) return DefWindowProc( hWnd, Msg, wParam, lParam);

	switch(Msg) {
	case WM_ERASEBKGND:
			// Disable background-erasing. It seems a little unreliable, so we'll
			// do it ourselves in WM_PAINT
		return 1;

	case WM_CONTEXTMENU: case WM_RBUTTONDOWN:
		ContextMenu(This,LOWORD(lParam),HIWORD(lParam), hWnd);
		return 0;

	case WM_PAINT:
		if (This->dib)
		{
			PAINTSTRUCT paintStruct;
			HDC hdc;
			RECT rect;
			
			hdc = BeginPaint(hWnd, &paintStruct);
			GetClientRect(hWnd, &rect);
			PaintBitmap(This, hdc, &rect);
			EndPaint(hWnd, &paintStruct);

			return 0;
		}
		else
			return This->fDefaultWindowProc(hWnd, Msg, wParam, lParam);
	}
	/* Forward unprocessed messages on to their original destination
	 * (the window proc we replaced) */
	return This->fDefaultWindowProc(hWnd, Msg, wParam, lParam);
}

LRESULT CALLBACK DlgProcAbout(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	switch(Msg) {
	case WM_INITDIALOG:
		{
			char buf[4096];
			char info[1024];
			DWORD tabs[1];

			PluginInstance *This=(PluginInstance*)lParam;

			tabs[0]= 60;
			SendDlgItemMessage(hWnd,IDC_IMGINFO,EM_SETTABSTOPS,(WPARAM)1,(LPARAM)tabs);

			sprintf(buf,"WebWKO Plug-in, Version %s\r\n%s"
#ifdef _DEBUG
				" DEBUG BUILD"
#endif
				"\r\nBy Daniel Vollmer (maven@maven.de),\r\nbased on WebBMP by Jason Summers."
				,VERS,__DATE__);
			SetDlgItemText(hWnd,IDC_PRGINFO,buf);

			sprintf(buf,"URL:\t%s\r\n",This->url);
				
			if (!This->valid)
				GetImageInfo(This);
			if (This->status == ST_LOADED)
			{
				sprintf(info,"Size:\t%d bytes\r\n", This->bits_read / 8);
				strcat(buf, info);
#ifdef _DEBUG
				sprintf(info, "Reported size:\t%d bytes\r\n", This->streamend);
				strcat(buf, info);
				sprintf(info, "Memory alloc:\t%d bytes\r\n", This->bytes_alloc);
				strcat(buf, info);
				sprintf(info, "Num blocks:\t%d\r\n", This->wvh.num_blocks);
				strcat(buf, info);
#endif
				sprintf(info, "Dimensions:\t%d x %d pixels\r\n", This->wvh.owidth, This->wvh.oheight);
				strcat(buf, info);

				sprintf(info, "# of channels:\t%d\r\n", This->wvh.num_channels);
				strcat(buf,info);
				
				sprintf(info, "YCbCr:\t%d\r\n", This->wvh_is_yuv);
				strcat(buf, info);
			}

			sprintf(info,"Browser:\t%s\r\n",NPN_UserAgent(This->instance));
			strcat(buf,info);

			sprintf(info,"Browser API:\t%d.%d\r\n",HIBYTE(g_pNavigatorFuncs->version),
				LOBYTE(g_pNavigatorFuncs->version));
			strcat(buf,info);

			sprintf(info,"Plug-in API:\t%d.%d\r\n",NP_VERSION_MAJOR,NP_VERSION_MINOR);
			strcat(buf,info);

			SetDlgItemText(hWnd,IDC_IMGINFO,buf);
		}
		return(TRUE);
	case WM_CLOSE:
		EndDialog(hWnd,0);
		return(TRUE);
	case WM_COMMAND:
		switch(wParam) {
		case IDOK:
		case IDCANCEL:
			EndDialog(hWnd,0);
			return(TRUE);
		}
	}
	return(FALSE);
}
